@@ -1,15 +1,19 @@
 """`matflow.utils.py`"""
 
+import os
 import sys
 import io
 import collections
 import copy
 import itertools
+import h5py
 import numpy as np
 import random
 import re
 import time
-from contextlib import redirect_stdout
+from contextlib import redirect_stdout, contextmanager
+from datetime import datetime
+from pathlib import Path
 
 from ruamel.yaml import YAML
 
@@ -134,22 +138,22 @@ def nest_lists(my_list):
 
 
 def repeat(lst, reps):
-    'Repeat 1D list elements.'
+    """Repeat 1D list elements."""
     return list(itertools.chain.from_iterable(itertools.repeat(x, reps) for x in lst))
 
 
 def tile(lst, tiles):
-    'Tile a 1D list.'
+    """Tile a 1D list."""
     return lst * tiles
 
 
 def index(lst, idx):
-    'Get elements of a list.'
+    """Get elements of a list."""
     return [lst[i] for i in idx]
 
 
 def arange(size):
-    'Get 1D list of increasing integers.'
+    """Get 1D list of increasing integers."""
     return list(range(size))
 
 
@@ -264,7 +268,7 @@ def dump_to_yaml_string(data):
 
 
 def get_specifier_dict(key, name_key=None, base_key=None, defaults=None,
-                       list_specifiers=None):
+                       list_specifiers=None, cast_types=None):
     """Resolve a string key with additional specifiers using square-brackets into a dict.
 
     Parameters
@@ -275,6 +279,8 @@ def get_specifier_dict(key, name_key=None, base_key=None, defaults=None,
     defaults : dict
     list_specifiers : list of str
         Any specifier in this list will be added to the returned dict as a list element.
+    cast_types : dict
+        Dict of (key: type) to cast those keys' values to.
 
     Returns
     -------
@@ -298,6 +304,7 @@ def get_specifier_dict(key, name_key=None, base_key=None, defaults=None,
     """
 
     list_specifiers = list_specifiers or []
+    cast_types = cast_types or {}
     out = {}
 
     if isinstance(key, str):
@@ -351,6 +358,14 @@ def get_specifier_dict(key, name_key=None, base_key=None, defaults=None,
         if k not in out:
             out[k] = copy.deepcopy(v)
 
+    for key, cast_type in cast_types.items():
+        if key in out:
+            if cast_type is bool:
+                new_val = cast_bool(out[key])
+            else:
+                new_val = cast_type(out[key])
+            out[key] = new_val
+
     return out
 
 
@@ -377,3 +392,118 @@ def extract_variable_names(source_str, delimiters):
     var_names = re.findall(pattern, source_str)
 
     return var_names
+
+
+def get_nested_item(obj, address):
+    out = obj
+    for i in address:
+        out = out[i]
+    return out
+
+
+def get_workflow_paths(base_dir, quiet=True):
+    base_dir = Path(base_dir)
+    wkflows = []
+    for i in base_dir.glob('**/*'):
+        if i.name == 'workflow.hdf5':
+            wk_full_path = i
+            wk_rel_path = wk_full_path.relative_to(base_dir)
+            wk_disp_path = wk_rel_path.parent
+            with h5py.File(wk_full_path, 'r') as handle:
+                try:
+                    try:
+                        handle["/workflow_obj/data/'figures'"]
+                    except KeyError:
+                        if not quiet:
+                            print(f'No "figures" key for workflow: {wk_disp_path}.')
+                        continue
+                    timestamp_path = "/workflow_obj/data/'history'/data/data_0/'timestamp'/data"
+                    timestamp_dict = {k[1:-1]: v['data'][()]
+                                      for k, v in handle[timestamp_path].items()}
+                    timestamp = datetime(**timestamp_dict)
+                    wkflows.append({
+                        'ID': handle.attrs['workflow_id'],
+                        'full_path': str(wk_full_path),
+                        'display_path': str(wk_disp_path),
+                        'timestamp': timestamp,
+                        'display_timestamp': timestamp.strftime(r'%Y-%m-%d %H:%M:%S'),
+                    })
+                except:
+                    if not quiet:
+                        print(f'No timestamp for workflow: {wk_disp_path}')
+    return wkflows
+
+
+def order_workflow_paths_by_date(workflow_paths):
+    return sorted(workflow_paths, key=lambda x: x['timestamp'])
+
+
+def nested_dict_arrays_to_list(obj):
+    if isinstance(obj, np.ndarray):
+        obj = obj.tolist()
+    elif isinstance(obj, dict):
+        for key, val in obj.items():
+            obj[key] = nested_dict_arrays_to_list(val)
+    return obj
+
+
+def move_element_forward(lst, index, position, return_map=True):
+    """Move a list element forward in the list to a new index position."""
+
+    if index > position:
+        raise ValueError('`index` cannot be larger than `position`, since that would '
+                         'not be a "forward" move!')
+
+    if position > len(lst) - 1:
+        raise ValueError('`position` must be a valid list index.')
+
+    sub_list_1 = lst[:position + 1]
+    sub_list_2 = lst[position + 1:]
+    elem = sub_list_1.pop(index)
+    out = sub_list_1 + [elem] + sub_list_2
+
+    # Indices to the left of the element that is to be moved do not change:
+    idx_map_left = {i: i for i in range(0, index)}
+
+    # The index of the moved element changes to `position`
+    idx_map_element = {index: position}
+
+    # Indicies to the right of the element up to the new position are decremented:
+    idx_map_middle = {i: i - 1 for i in range(index + 1, position + 1)}
+
+    # Indices to the right of the new position do not change:
+    idx_map_right = {i: i for i in range(position + 1, len(lst))}
+
+    idx_map = {
+        **idx_map_left,
+        **idx_map_element,
+        **idx_map_middle,
+        **idx_map_right
+    }
+
+    if return_map:
+        return out, idx_map
+    else:
+        return out
+
+
+def cast_bool(bool_str):
+    if isinstance(bool_str, bool):
+        return bool_str
+    elif bool_str.lower() == 'true':
+        return True
+    elif bool_str.lower() == 'false':
+        return False
+    else:
+        raise ValueError(f'"{bool_str}" cannot be cast to True or False.')
+
+
+@contextmanager
+def working_directory(path):
+    """Change to a working directory and return to previous working directory on exit."""
+    prev_cwd = Path.cwd()
+    os.chdir(path)
+    try:
+        yield
+    finally:
+        os.chdir(prev_cwd)
